!$Id: infotrac.F90 5183 2024-09-10 15:14:23Z dcugnet $
!
MODULE infotrac

   USE       strings_mod, ONLY: msg, maxlen, cat, dispTable, int2str, bool2str, strStack, strParse, strCount, strIdx
   USE readTracFiles_mod, ONLY: readTracersFiles, maxTableWidth,  tisot=>isot_type, addPhase, addKey, iH2O, &
        indexUpdate, keys_type, testTracersFiles, processIsotopes, trac=>tracers,   delPhase, getKey, tran0
   USE readTracFiles_mod, ONLY: new2oldH2O, newHNO3, oldHNO3

   IMPLICIT NONE

   PRIVATE

   !=== FOR TRACERS:
   PUBLIC :: init_infotrac                                 !--- Initialization of the tracers
   PUBLIC :: tracers, type_trac                            !--- Full tracers database, tracers type keyword
   PUBLIC :: nqtot,   nbtr,   nqo,   nqCO2,   nqtottr      !--- Main dimensions
   PUBLIC :: conv_flg, pbl_flg                             !--- Convection & boundary layer activation keys
   PUBLIC :: new2oldH2O, newHNO3, oldHNO3                  !--- For backwards compatibility in dynetat0
   PUBLIC :: addPhase, delPhase                            !--- Add/remove the phase from the name of a tracer

   !=== FOR ISOTOPES: General
   PUBLIC :: isot_type, nbIso                              !--- Derived type, full isotopes families database + nb of families
   PUBLIC :: isoSelect, ixIso, isoFamilies                 !--- Isotopes families selection tool + selected index + list
   !=== FOR ISOTOPES: Specific to water
   PUBLIC :: iH2O                                          !--- Value of "ixIso" for "H2O" isotopes class
   PUBLIC :: min_qParent, min_qMass, min_ratio             !--- Min. values for various isotopic quantities
   !=== FOR ISOTOPES: Depending on the selected isotopes family
   PUBLIC :: isotope                                       !--- Selected isotopes database (argument of getKey)
   PUBLIC :: isoKeys, isoName, isoZone, isoPhas            !--- Isotopes keys & names, tagging zones names, phases
   PUBLIC ::    niso,   ntiso,   nzone,   nphas            !--- Number of   "   "
   PUBLIC :: itZonIso                                      !--- index "it" in "isoName(1:niso)" = f(tagging idx, isotope idx)
   PUBLIC :: iqIsoPha                                      !--- index "iq" in "qx"              = f(isotope idx,   phase idx)
   PUBLIC :: isoCheck                                      !--- Run isotopes checking routines
   !=== FOR BOTH TRACERS AND ISOTOPES
   PUBLIC :: getKey                                        !--- Get a key from "tracers" or "isotope"

!=== CONVENTIONS FOR TRACERS NUMBERS:
!  |--------------------+-----------------------+-----------------+---------------+----------------------------|
!  | water in different |    water tagging      |  water isotopes | other tracers | additional tracers moments |
!  | phases: H2O_[glsrb]|      isotopes         |                 |               |  for higher order schemes  |
!  |--------------------+-----------------------+-----------------+---------------+----------------------------|
!  |                    |                       |                 |               |                            |
!  |<--     nqo      -->|<-- nqo*niso* nzone -->|<-- nqo*niso  -->|<--  nbtr   -->|<--        (nmom)        -->|         
!  |                    |                                         |                                            |
!  |                    |<-- nqo*niso*(nzone+1)  =   nqo*ntiso -->|<--    nqtottr = nbtr + nmom             -->|
!  |                                                                              = nqtot - nqo*(ntiso+1)      |
!  |                                                                                                           |
!  |<--                        nqtrue  =  nbtr + nqo*(ntiso+1)                 -->|                            |
!  |                                                                                                           |
!  |<--                        nqtot   =  nqtrue + nmom                                                     -->|
!  |                                                                                                           |
!  |-----------------------------------------------------------------------------------------------------------|
!  NOTES FOR THIS TABLE:
!  * Used "niso", "nzone" and "ntiso" are components of "isotopes(ip)" for water (isotopes(ip)%name == 'H2O'),
!    since water is so far the sole tracers family, except passive CO2, removed from the main tracers table.
!  * For water, "nqo" is equal to the more general field "isotopes(ip)%nphas".
!  * "niso", "nzone", "ntiso", "nphas" are defined for other isotopic tracers families, if any.
!  * If you deal with an isotopes family other than "H2O" ("Sulf" in the example), a good practice is to keep
!    track of the isotopes class (of its index) before switching to it at the beginning of the dedicated code:
!  - first time (use selection by name and compute the corresponding index iSulf) :
!  i0=ixIso; IF(.NOT.isoSelect('Sulf')) CALL abort_gcm("Can't select isotopes class", modname, 1); iS=ixIso
!  - next times (use selection by index - "iS" has been computed at first call):
!  i0=ixIso; IF(.NOT.isoSelect(iS))     CALL abort_gcm("Can't select isotopes class", modname, 1)
!    and to switch back to the original category when you're done with "Sulf":
!            IF(.NOT.isoSelect(i0))     CALL abort_gcm("Can't select isotopes class", modname, 1)
!    to restore the original isotopes category (before dealing with "Sulf" (most of the time "H2O").
!
!=== LOCAL DERIVED TYPE "trac_type" EMBEDDING MOST OF THE TRACERS-RELATED QUANTITIES (LENGTH: nqtot)
!    Each entry is accessible using "%" sign.
!  |-------------+------------------------------------------------------+-------------+------------------------+
!  |  entry      | Meaning                                              | Former name | Possible values        |
!  |-------------+------------------------------------------------------+-------------+------------------------+
!  | name        | Name (short)                                         | tname       |                        |
!  | keys        | key/val pairs accessible with "getKey" routine       | /           |                        |
!  | gen0Name    | Name of the 1st generation ancestor                  | /           |                        |
!  | parent      | Name of the parent                                   | /           |                        |
!  | longName    | Long name (with adv. scheme suffix) for outputs      | ttext       |                        |
!  | type        | Type (so far: tracer or tag)                         | /           | tracer,tag             |
!  | phase       | Phases list ("g"as / "l"iquid / "s"olid              |             | [g|l|s|r|b]            |
!  |             |              "r"(cloud) / "b"lowing)                 | /           |                        |
!  | component   | Name(s) of the merged/cumulated section(s)           | /           | coma-separated names   |
!  | iGeneration | Generation (>=1)                                     | /           |                        |
!  | iqParent    | Index of the parent tracer                           | iqpere      | 1:nqtot                |
!  | iqDescen    | Indexes of the childs       (all generations)        | iqfils      | 1:nqtot                |
!  | nqDescen    | Number of the descendants   (all generations)        | nqdesc      | 1:nqtot                |
!  | nqChildren  | Number of childs            (1st generation only)    | nqfils      | 1:nqtot                |
!  | iadv        | Advection scheme number                              | iadv        | 1,2,10-20(exc.15,19),30|
!  | iso_iGroup  | Isotopes group index in isotopes(:)                  | /           | 1:nbIso                |
!  | iso_iName   | Isotope  name  index in isotopes(iso_iGroup)%trac(:) | iso_indnum  | 1:niso                 |
!  | iso_iZone   | Isotope  zone  index in isotopes(iso_iGroup)%zone(:) | zone_num    | 1:nzone                |
!  | iso_iPhas   | Isotope  phase index in isotopes(iso_iGroup)%phas(:) | phase_num   | 1:nphas                |
!  +-------------+------------------------------------------------------+-------------+------------------------+
!
!=== DERIVED TYPE EMBEDDING MOST OF THE ISOTOPES-RELATED QUANTITIES (LENGTH: nbIso, NUMBER OF ISOTOPES FAMILIES)
!    Each entry is accessible using "%" sign.
!  |-----------------+--------------------------------------------------+--------------------+-----------------+
!  |  entry | length | Meaning                                          |    Former name     | Possible values |
!  |-----------------+--------------------------------------------------+--------------------+-----------------+
!  | name            | Name of the isotopes class (family)              |                    |                 |
!  | keys   | niso   | Isotopes keys/values pairs list + number         |                    |                 |
!  | trac   | ntiso  | Isotopes + tagging tracers list + number         | / | ntraciso       |                 |
!  | zone   | nzone  | Geographic tagging zones   list + number         | / | ntraceurs_zone |                 |
!  | phase  | nphas  | Phases                     list + number         |                    | [g|l|s|r|b] 1:5 |
!  | iqIsoPha        | Index in "qx"           = f(name(1:ntiso)),phas) | iqiso              | 1:nqtot         |
!  | itZonIso        | Index in "trac(1:ntiso)"= f(zone, name(1:niso))  | index_trac         | 1:ntiso         |
!  +-----------------+--------------------------------------------------+--------------------+-----------------+

!------------------------------------------------------------------------------------------------------------------------------
   TYPE :: trac_type                                            !=== TYPE FOR A SINGLE TRACER NAMED "name"
     CHARACTER(LEN=maxlen) :: name        = ''                  !--- Name of the tracer
     TYPE(keys_type)       :: keys                              !--- <key>=<val> pairs vector (general container)
     CHARACTER(LEN=maxlen) :: gen0Name    = ''                  !--- First generation ancestor name
     CHARACTER(LEN=maxlen) :: parent      = ''                  !--- Parent name
     CHARACTER(LEN=maxlen) :: longName    = ''                  !--- Long name (with advection scheme suffix)
     CHARACTER(LEN=maxlen) :: type        = 'tracer'            !--- Type  (so far: 'tracer' / 'tag')
     CHARACTER(LEN=maxlen) :: phase       = 'g'                 !--- Phase ('g'as / 'l'iquid / 's'olid)
     CHARACTER(LEN=maxlen) :: component   = ''                  !--- Coma-separated list of components (Ex: lmdz,inca)
     INTEGER               :: iGeneration = -1                  !--- Generation number (>=0)
     INTEGER               :: iqParent    = 0                   !--- Parent index
     INTEGER,  ALLOCATABLE :: iqDescen(:)                       !--- Descendants index (in growing generation order)
     INTEGER               :: nqDescen    = 0                   !--- Number of descendants (all generations)
     INTEGER               :: nqChildren  = 0                   !--- Number of children  (first generation)
     INTEGER               :: iadv        = 10                  !--- Advection scheme used
     INTEGER               :: iso_iGroup  = 0                   !--- Isotopes group index in isotopes(:)
     INTEGER               :: iso_iName   = 0                   !--- Isotope  name  index in isotopes(iso_iGroup)%trac(:)
     INTEGER               :: iso_iZone   = 0                   !--- Isotope  zone  index in isotopes(iso_iGroup)%zone(:)
     INTEGER               :: iso_iPhase  = 0                   !--- Isotope  phase index in isotopes(iso_iGroup)%phase
   END TYPE trac_type
!------------------------------------------------------------------------------------------------------------------------------
  TYPE :: isot_type                                             !=== TYPE FOR THE ISOTOPES FAMILY DESCENDING ON TRACER "name"
    CHARACTER(LEN=maxlen)              :: name                  !--- Isotopes family name (ex: H2O)
    TYPE(keys_type),       ALLOCATABLE :: keys(:)               !--- Isotopes keys/values pairs list     (length: niso)
    LOGICAL                            :: check=.FALSE.         !--- Flag for checking routines triggering
    CHARACTER(LEN=maxlen), ALLOCATABLE :: trac(:)               !--- Isotopes + tagging tracers list     (length: ntiso)
    CHARACTER(LEN=maxlen), ALLOCATABLE :: zone(:)               !--- Geographic tagging zones names list (length: nzone)
    CHARACTER(LEN=maxlen)              :: phase = 'g'           !--- Phases list: [g|l|s|r|b]            (length: nphas)
    INTEGER                            :: niso  = 0             !--- Number of isotopes, excluding tagging tracers
    INTEGER                            :: ntiso = 0             !--- Number of isotopes, including tagging tracers
    INTEGER                            :: nzone = 0             !--- Number of geographic tagging zones
    INTEGER                            :: nphas = 0             !--- Number of phases
    INTEGER,               ALLOCATABLE :: iqIsoPha(:,:)         !--- Idx in "tracers(1:nqtot)" = f(     name(1:ntiso) ,phas)
    INTEGER,               ALLOCATABLE :: itZonIso(:,:)         !--- Idx in "trac(1:ntiso)"    = f(zone,name(1:niso))
  END TYPE isot_type
!------------------------------------------------------------------------------------------------------------------------------
  INTERFACE isoSelect; MODULE PROCEDURE isoSelectByIndex, isoSelectByName; END INTERFACE isoSelect
!------------------------------------------------------------------------------------------------------------------------------

   !=== THRESHOLDS FOR WATER
   REAL, PARAMETER :: min_qParent = 1.e-30, min_qMass = 1.e-18, min_ratio = 1.e-16 ! MVals et CRisi

   !=== DIMENSIONS OF THE TRACERS TABLES AND OTHER SCALAR VARIABLES
   INTEGER,               SAVE :: nqtot,   &                    !--- Tracers nb in dynamics (incl. higher moments + H2O)
                                  nbtr,    &                    !--- Tracers nb in physics  (excl. higher moments + H2O)
                                  nqo,     &                    !--- Number of water phases
                                  nqtottr, &                    !--- Number of tracers passed to phytrac (TO BE DELETED ?)
                                  nqCO2                         !--- Number of tracers of CO2  (ThL)
   CHARACTER(LEN=maxlen), SAVE :: type_trac                     !--- Keyword for tracers type(s)

   !=== NUMBER AND LIST OF DEFINED ISOTOPES FAMILIES
   INTEGER,               SAVE              :: nbIso            !--- Number of defined isotopes classes
   CHARACTER(LEN=maxlen), SAVE, ALLOCATABLE :: isoFamilies(:)   !--- Generation 0 tracer name for each isotopes family (nbIso)

   !=== QUANTITIES RELATED TO THE CURRENTLY SELECTED ISOTOPES CLASS (USUALLY H2O)
   TYPE(isot_type),       SAVE, POINTER :: isotope              !--- Selected isotopes database (=isotopes(ixIso))
   TYPE(keys_type),       SAVE, POINTER :: isoKeys(:)           !--- Database to get isotopes keys using "getKey"       (niso)
   CHARACTER(LEN=maxlen), SAVE, POINTER :: isoName(:),      &   !--- Isotopes list including tagging tracers, no phase (ntiso)
                                           isoZone(:),      &   !--- Geographic tagging zones list                     (nzone)
                                           isoPhas              !--- Used phases names ([g|l|s|r|b])                   (nphas)
   INTEGER,               SAVE, POINTER :: itZonIso(:,:),   &   !--- Idx "it" in isoName(1:niso) = f(tagging idx, isotope idx)
                                           iqIsoPha(:,:)        !--- Idx "iq" in qx              = f(isotope idx,   phase idx)
   INTEGER,               SAVE          :: ixIso,           &   !--- Idx in "isoFamilies" of currently selectd class
                                           niso,            &   !--- Number of isotopes
                                           ntiso,           &   !--- Number of isotopes + tagging tracers
                                           nzone,           &   !--- Number of tagging zones
                                           nphas                !--- Number of phases
   LOGICAL,               SAVE          :: isoCheck             !--- Isotopes checking routines triggering flag

   !=== VARIABLES FOR INCA
   INTEGER, DIMENSION(:), SAVE, ALLOCATABLE :: &
                    conv_flg, pbl_flg                           !--- Convection / boundary layer activation (nbtr)

   !=== TRACERS/ISOTOPES DESCRIPTORS: EFFECTIVE STORAGE (LOCAL DERIVED TYPES)
   TYPE(trac_type), SAVE, ALLOCATABLE, TARGET ::  tracers(:)
   TYPE(isot_type), SAVE, ALLOCATABLE, TARGET :: isotopes(:)

CONTAINS

SUBROUTINE init_infotrac
   USE control_mod, ONLY: planet_type
#ifdef CPP_IOIPSL
    USE IOIPSL, ONLY: getin
#else
    USE ioipsl_getincom, only: getin
#endif
#ifdef CPP_PARA
   USE parallel_lmdz, ONLY: is_master
#endif
#ifdef REPROBUS
   USE CHEM_REP, ONLY: Init_chem_rep_trac
#endif
   IMPLICIT NONE
!==============================================================================================================================
!
!   Auteur:  P. Le Van /L. Fairhead/F.Hourdin
!   -------
!
!   Modifications:
!   --------------
!   05/94: F.Forget      Modif special traceur
!   02/02: M-A Filiberti Lecture de traceur.def
!   01/22: D. Cugnet     Nouveaux tracer.def et tracer_*.def + encapsulation (types trac_type et isot_type)
!
!   Objet:
!   ------
!   GCM LMD nouvelle grille
!
!==============================================================================================================================
!   ... modification de l'integration de q ( 26/04/94 ) ....
!------------------------------------------------------------------------------------------------------------------------------
! Declarations:
   INCLUDE "dimensions.h"
   INCLUDE "iniprint.h"

!------------------------------------------------------------------------------------------------------------------------------
! Local variables
   INTEGER, ALLOCATABLE :: hadv(:), vadv(:), itmp(:)                 !--- Horizontal/vertical transport scheme number
#ifdef INCA
   INTEGER, ALLOCATABLE :: had (:), hadv_inca(:), conv_flg_inca(:), &!--- Variables specific to INCA
                           vad (:), vadv_inca(:),  pbl_flg_inca(:)
   CHARACTER(LEN=8), ALLOCATABLE :: solsym_inca(:)                   !--- Tracers names for INCA
   INTEGER :: nqINCA
#endif
#ifndef CPP_PARA
   LOGICAL               :: is_master=.TRUE.
#endif
   CHARACTER(LEN=2)      ::   suff(9)                                !--- Suffixes for schemes of order 3 or 4 (Prather)
   CHARACTER(LEN=3)      :: descrq(30)                               !--- Advection scheme description tags
   CHARACTER(LEN=maxlen) :: msg1, texp, ttp, ky, nam, val            !--- Strings for messages and expanded tracers type
   INTEGER :: fType                                                  !--- Tracers description file type ; 0: none
                                                                     !--- 1/2/3: "traceur.def"/"tracer.def"/"tracer_*.def"
   INTEGER :: nqtrue                                                 !--- Tracers nb from tracer.def (no higher order moments)
   INTEGER :: iad                                                    !--- Advection scheme number
   INTEGER :: iq, jq, it, nt, im, nm, ig                             !--- Indexes and temporary variables
   LOGICAL :: lerr, lInit
   TYPE(keys_type), ALLOCATABLE, TARGET :: tra(:)                    !--- Tracers  descriptor as in readTracFiles_mod
   TYPE(tisot),     ALLOCATABLE         :: iso(:)                    !--- Isotopes descriptor as in readTracFiles_mod
   TYPE(trac_type), ALLOCATABLE, TARGET :: ttr(:)
   TYPE(trac_type), POINTER             :: t(:), t1
   TYPE(keys_type), POINTER             :: k(:)
   CHARACTER(LEN=maxlen),   ALLOCATABLE :: types_trac(:)             !--- Keywords for tracers type(s), parsed version
   CHARACTER(LEN=*), PARAMETER :: modname="init_infotrac"
!------------------------------------------------------------------------------------------------------------------------------
! Initialization :
!------------------------------------------------------------------------------------------------------------------------------
   suff          = ['x ','y ','z ','xx','xy','xz','yy','yz','zz']
   descrq( 1:30) =  '   '
   descrq( 1: 2) = ['LMV','BAK']
   descrq(10:20) = ['VL1','VLP','FH1','FH2','VLH','   ','PPM','PPS','PPP','   ','SLP']
   descrq(30)    =  'PRA'

   CALL msg('type_trac = "'//TRIM(type_trac)//'"', modname, is_master)
   IF(strCount(type_trac, '|', nt)) CALL abort_gcm(modname, 'Could''nt parse the "type_trac" string with delimiter "|"', 1)
   IF(nt >= 3) CALL abort_gcm(modname, 'you need to modify type_trac, this version is not supported by lmdz', 1)
   IF(strParse(type_trac, '|', types_trac, n=nt)) CALL abort_gcm(modname, "couldn't parse "//'"type_trac"', 1)
   IF(nt == 2) type_trac = types_trac(2) ! TO BE DELETED SOON

   lInit = .NOT.ALLOCATED(trac)

!##############################################################################################################################
   IF(lInit .AND. is_master) THEN                                    !=== SKIPED IF ALREADY DONE
!##############################################################################################################################
   !--- MESSAGE ABOUT THE CHOSEN CONFIGURATION
   msg1 = 'For type_trac = "'//TRIM(type_trac)//'":'
   SELECT CASE(type_trac)
      CASE('inca'); CALL msg(TRIM(msg1)//' coupling with INCA chemistry model',        modname)
      CASE('inco'); CALL msg(TRIM(msg1)//' coupling jointly with INCA and CO2 cycle',  modname)
      CASE('repr'); CALL msg(TRIM(msg1)//' coupling with REPROBUS chemistry model',    modname)
      CASE('co2i'); CALL msg(TRIM(msg1)//' you have chosen to run with CO2 cycle',     modname)
      CASE('coag'); CALL msg(TRIM(msg1)//' tracers are treated for COAGULATION tests', modname)
      CASE('lmdz'); CALL msg(TRIM(msg1)//' tracers are treated in LMDZ only',          modname)
      CASE DEFAULT; CALL abort_gcm(modname,'type_trac='//TRIM(type_trac)//' not possible yet.',1)
   END SELECT

   !--- COHERENCE TEST BETWEEN "type_trac" AND PREPROCESSING KEYS
   SELECT CASE(type_trac)
      CASE('inca', 'inco')
#ifndef INCA
         CALL abort_gcm(modname, 'You must add cpp key INCA and compile with INCA code', 1)
#endif
      CASE('repr')
#ifndef REPROBUS
         CALL abort_gcm(modname, 'You must add cpp key REPROBUS and compile with REPROBUS code', 1)
#endif
      CASE('coag')
#ifndef CPP_StratAer
         CALL abort_gcm(modname, 'You must add cpp key StratAer and compile with StratAer code', 1)
#endif
   END SELECT
!##############################################################################################################################
   END IF
!##############################################################################################################################

!==============================================================================================================================
! 0) DETERMINE THE TYPE OF THE INPUT TRACERS DESCRIPTION FILE AND READ IT ; TRANSFER THE NEEDED QUANTITIES TO LOCAL "tracers".
!==============================================================================================================================
   texp = type_trac                                                            !=== EXPANDED (WITH "|" SEPARATOR) "type_trac"
   IF(texp == 'inco') texp = 'co2i|inca'
   IF(texp /= 'lmdz') texp = 'lmdz|'//TRIM(texp)
   IF(testTracersFiles(modname, texp, fType, lInit.AND.is_master)) CALL abort_gcm(modname, 'problem with tracers file(s)',1)
   ttp = type_trac; IF(fType /= 1) ttp = texp
   !---------------------------------------------------------------------------------------------------------------------------
   IF(fType == 0) CALL abort_gcm(modname, 'Missing "traceur.def", "tracer.def" or "tracer_<keyword>.def tracers file.',1)
   !---------------------------------------------------------------------------------------------------------------------------
   IF(fType == 1 .AND. ANY(['inca', 'inco'] == type_trac) .AND. lInit) &       !=== FOUND OLD STYLE INCA "traceur.def"
      CALL abort_gcm(modname, 'retro-compatibility with old-style INCA traceur.def files has been disabled.', 1)
   !---------------------------------------------------------------------------------------------------------------------------

!##############################################################################################################################
   IF(lInit) THEN
      IF(readTracersFiles(ttp, tra, type_trac == 'repr')) CALL abort_gcm(modname, 'problem with tracers file(s)',1)
   ELSE
      tra = trac
   END IF
   CALL msg('No tracers description file(s) reading needed: already done', modname, .NOT.lInit.AND.is_master)
!##############################################################################################################################

   !--- POPULATE SOME EXPLICIT (ACCESSIBLE THROUGH "%") KEYS OF THE LOCAL TRACERS DESCRIPTION DERIVED TYPE
   !    To be defined: iqParent, iq/nqDescen, nqChildren (in indexUpdate), longName, iso_i*, iadv (later)
   ALLOCATE(tracers(SIZE(tra)))
   DO iq = 1, SIZE(tra); t1 => tracers(iq)
      t1%keys = tra(iq)
      msg1 = '" for tracer nr. '//TRIM(int2str(iq))
      ky='name       '; IF(getKey(ky, t1%name,        iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
      msg1 = '" for "'//TRIM(t1%name)//'"'
      ky='gen0Name   '; IF(getKey(ky, t1%gen0Name,    iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='parent     '; IF(getKey(ky, t1%parent,      iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='type       '; IF(getKey(ky, t1%type,        iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='phase      '; IF(getKey(ky, t1%phase,       iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='component  '; IF(getKey(ky, t1%component,   iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='iGeneration'; IF(getKey(ky, t1%iGeneration, iq, tra)) CALL abort_gcm(modname, 'No key "'//TRIM(ky)//TRIM(msg1), 1)
   END DO

!==============================================================================================================================
! 1) Get various numbers: "nqtrue" (first order only tracers), "nqo" (water phases), 'nbtr' (tracers passed to physics), etc.
!==============================================================================================================================
   nqtrue = SIZE(tracers)                                                                               !--- "true" tracers
   nqo    =      COUNT(tracers(:)%component == 'lmdz' .AND. delPhase(tracers(:)%name)     == 'H2O')     !--- Water phases
   nbtr = nqtrue-COUNT(tracers(:)%component == 'lmdz' .AND. delPhase(tracers(:)%gen0Name) == 'H2O')     !--- Passed to phytrac
   nqCO2  =      COUNT( [type_trac == 'inco', type_trac == 'co2i'] )
#ifdef INCA
   nqINCA =      COUNT(tracers(:)%component == 'inca')
#endif
#ifdef REPROBUS
   CALL Init_chem_rep_trac(nbtr, nqo, tracers(:)%name)                         !--- Transfert the number of tracers to Reprobus
#endif

!==============================================================================================================================
! 2) Calculate nqtot, number of tracers needed (greater if advection schemes 20 or 30 have been chosen).
!==============================================================================================================================
   IF(getKey('hadv', hadv, ky=tra)) CALL abort_gcm(modname, 'missing key "hadv"', 1)
   IF(getKey('vadv', vadv, ky=tra)) CALL abort_gcm(modname, 'missing key "vadv"', 1)
   DO iq = 1, nqtrue
      IF( hadv(iq)<20 .OR. (ANY(hadv(iq)==[20,30]) .AND. hadv(iq)==vadv(iq)) ) CYCLE
      WRITE(msg1,'("The choice hadv=",i0,", vadv=",i0,a)')hadv(iq),vadv(iq),' for "'//TRIM(tracers(iq)%name)//'" is not available'
      CALL abort_gcm(modname, TRIM(msg1), 1)
   END DO
   nqtot =    COUNT( hadv< 20 .AND. vadv< 20 ) &                     !--- No additional tracer
         +  4*COUNT( hadv==20 .AND. vadv==20 ) &                     !--- 3  additional tracers
         + 10*COUNT( hadv==30 .AND. vadv==30 )                       !--- 9  additional tracers

   !--- More tracers due to the choice of advection scheme => assign total number of tracers
   IF( nqtot /= nqtrue ) THEN
      CALL msg('The choice of advection scheme for one or more tracers makes it necessary to add tracers')
      CALL msg('The number of true tracers is '//TRIM(int2str(nqtrue)))
      CALL msg('The total number of tracers needed is '//TRIM(int2str(nqtot)))
   END IF

!==============================================================================================================================
! 3) Determine the advection scheme choice for water and tracers "iadv" and the fields long name.
!     iadv = 1    "LMDZ-specific humidity transport" (for H2O vapour)          LMV
!     iadv = 2    backward                           (for H2O liquid)          BAK
!     iadv = 14   Van-Leer + specific humidity, modified by Francis Codron     VLH
!     iadv = 10   Van-Leer (chosen for vapour and liquid water)                VL1
!     iadv = 11   Van-Leer for hadv and PPM version (Monotonic) for vadv       VLP
!     iadv = 12   Frederic Hourdin I                                           FH1
!     iadv = 13   Frederic Hourdin II                                          FH2
!     iadv = 16   Monotonic         PPM (Collela & Woodward 1984)              PPM
!     iadv = 17   Semi-monotonic    PPM (overshoots allowed)                   PPS
!     iadv = 18   Definite positive PPM (overshoots and undershoots allowed)   PPP
!     iadv = 20   Slopes                                                       SLP
!     iadv = 30   Prather                                                      PRA
!
!        In array q(ij,l,iq) : iq = 1/2[/3]    for vapour/liquid[/ice] water
!        And optionaly:        iq = 3[4],nqtot for other tracers
!==============================================================================================================================
   ALLOCATE(ttr(nqtot))
   jq = nqtrue+1
   DO iq = 1, nqtrue
      t1 => tracers(iq)

      !--- VERIFY THE CHOICE OF ADVECTION SCHEME
      iad = -1
      IF(hadv(iq)     ==    vadv(iq)    ) iad = hadv(iq)
      IF(hadv(iq)==10 .AND. vadv(iq)==16) iad = 11
      WRITE(msg1,'("Bad choice of advection scheme for ",a,": hadv = ",i0,", vadv = ",i0)')TRIM(t1%name), hadv(iq), vadv(iq)
      IF(iad == -1) CALL abort_gcm(modname, msg1, 1)

      !--- SET FIELDS longName, iadv
      t1%iadv       = iad
      t1%longName   = t1%name; IF(iad > 0) t1%longName=TRIM(t1%name)//descrq(iad)
      ttr(iq)       = t1

      !--- DEFINE THE HIGHER ORDER TRACERS, IF ANY
      nm = 0
      IF(iad == 20) nm = 3                                           !--- 2nd order scheme
      IF(iad == 30) nm = 9                                           !--- 3rd order scheme
      IF(nm == 0) CYCLE                                              !--- No higher moments
      ttr(jq+1:jq+nm)             = t1
      ttr(jq+1:jq+nm)%name        = [ (TRIM(t1%name)    //'-'//TRIM(suff(im)), im=1, nm) ]
      ttr(jq+1:jq+nm)%gen0Name    = [ (TRIM(t1%name)    //'-'//TRIM(suff(im)), im=1, nm) ]
      ttr(jq+1:jq+nm)%parent      = [ (TRIM(t1%parent)  //'-'//TRIM(suff(im)), im=1, nm) ]
      ttr(jq+1:jq+nm)%longName    = [ (TRIM(t1%longName)//'-'//TRIM(suff(im)), im=1, nm) ]
      ttr(jq+1:jq+nm)%iadv        = [ (-iad,    im=1, nm) ]
      jq = jq + nm
   END DO
   DEALLOCATE(hadv, vadv)
   CALL MOVE_ALLOC(FROM=ttr, TO=tracers)

   !--- SET FIELDS iqParent, iqDescen, nqDescen, nqChildren
   IF(indexUpdate(tracers%keys)) CALL abort_gcm(modname, 'problem with tracers indices update', 1)
   k => tracers(:)%keys
   DO iq = 1, SIZE(tracers); t1 => tracers(iq); msg1 = '" for "'//TRIM(t1%name)//'"'
      ky='iqParent  '; IF(getKey(ky, t1%iqParent,   iq, k)) CALL abort_gcm(modname, 'missing key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='iqDescen  '; IF(getKey(ky, t1%iqDescen,   iq, k)) CALL abort_gcm(modname, 'missing key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='nqDescen  '; IF(getKey(ky, t1%nqDescen,   iq, k)) CALL abort_gcm(modname, 'missing key "'//TRIM(ky)//TRIM(msg1), 1)
      ky='nqChildren'; IF(getKey(ky, t1%nqChildren, iq, k)) CALL abort_gcm(modname, 'missing key "'//TRIM(ky)//TRIM(msg1), 1)
   END DO

   !=== TEST ADVECTION SCHEME
   DO iq = 1, nqtot ; t1 => tracers(iq)
      iad = t1%iadv
      ig  = t1%iGeneration
      nam = t1%name
      val = 'iadv='//TRIM(int2str(iad))

      !--- ONLY TESTED VALUES FOR TRACERS FOR NOW:               iadv = 14, 10 (and 0 for non-transported tracers)
      IF(ALL([10,14,0] /= iad)) CALL abort_gcm(modname, TRIM(val)//' has not been tested yet ; 10 or 14 only are allowed !', 1)

      !--- ONLY TESTED VALUES SO FAR FOR PARENTS HAVING CHILDREN: iadv = 14, 10 (PARENTS: TRACERS OF GENERATION 0)
      IF(ALL([10,14] /= iad) .AND. ig == 0 .AND. ANY(tracers(:)%parent==nam)) &
         CALL abort_gcm(modname, TRIM(val)//' is not implemented for parents ; 10 or 14 only are allowed !', 1)

      !--- ONLY TESTED VALUES SO FAR FOR DESCENDANTS (TRACERS OF GENERATION > 0): iadv = 10 ; WATER VAPOUR: iadv = 14
      lerr = iad /= 10 .AND. ig > 0;                     IF(lerr) tracers(iq)%iadv = 10
      CALL msg('WARNING! '//TRIM(val)//  ' not implemented for children. Setting iadv=10 for "'//TRIM(nam)//'"', modname, lerr)
      lerr = iad == 14 .AND. nam /= addPhase('H2O','g'); IF(lerr) tracers(iq)%iadv = 10
      CALL msg('WARNING! '//TRIM(val)//' is valid for water vapour only. Setting iadv=10 for "'//TRIM(nam)//'"', modname, lerr)
   END DO

   !=== DETERMINE ISOTOPES RELATED PARAMETERS ; DEFINE THE EXPLICIT KEYS iso_i*
   IF(processIsotopes(tracers%keys, iso)) CALL abort_gcm(modname, 'problem while processing isotopes parameters', 1)

   !--- POPULATE SOME EXPLICIT (ACCESSIBLE THROUGH "%") KEYS OF THE LOCAL ISOTOPES DESCRIPTION DERIVED TYPE
   nbIso = SIZE(iso)
   ALLOCATE(isotopes(nbIso))
   IF(nbIso /= 0) THEN
      k => tracers(:)%keys
      IF(getKey('iso_iGroup', itmp, ky=k)) CALL abort_gcm(modname, 'missing key "iso_iGroup"', 1); tracers%iso_iGroup=itmp
      IF(getKey('iso_iName',  itmp, ky=k)) CALL abort_gcm(modname, 'missing key "iso_iName"',  1); tracers%iso_iName =itmp
      IF(getKey('iso_iZone',  itmp, ky=k)) CALL abort_gcm(modname, 'missing key "iso_iZone"',  1); tracers%iso_iZone =itmp
      IF(getKey('iso_iPhas',  itmp, ky=k)) CALL abort_gcm(modname, 'missing key "iso_iPhas"',  1); tracers%iso_iPhase=itmp
      isotopes(:)%name  = iso(:)%name                           !--- Isotopes family name (ex: H2O)
      isotopes(:)%phase = iso(:)%phase                          !--- Phases list: [g][l][s]              (length: nphas)
      isotopes(:)%niso  = iso(:)%niso                           !--- Number of isotopes, excluding tagging tracers
      isotopes(:)%ntiso = iso(:)%ntiso                          !--- Number of isotopes, including tagging tracers
      isotopes(:)%nzone = iso(:)%nzone                          !--- Number of geographic tagging zones
      isotopes(:)%nphas = iso(:)%nphas                          !--- Number of phases
      isotopes(:)%check = .FALSE.                               !--- Flag for checking routines triggering
      CALL getin('ok_iso_verif', isotopes(:)%check)
      DO it = 1, nbIso
         isotopes(it)%keys     = iso(it)%keys                   !--- Isotopes keys/values pairs list     (length: niso)
         isotopes(it)%trac     = iso(it)%trac                   !--- Isotopes + tagging tracers list     (length: ntiso)
         isotopes(it)%zone     = iso(it)%zone                   !--- Geographic tagging zones names list (length: nzone)
         isotopes(it)%iqIsoPha = iso(it)%iqIsoPha(:,:)          !--- Idx in "tracers(1:nqtot)" = f(name(1:ntiso)),phas)
         isotopes(it)%itZonIso = iso(it)%itZonIso(:,:)          !--- Idx in "tracers(1:ntiso)" = f(  zone,name(1:niso))
      END DO
      IF(isoSelect(1, .TRUE.)) CALL abort_gcm(modname, "Can't select the first isotopes family", 1)
      IF(.NOT.isoSelect('H2O', .TRUE.)) iH2O = ixIso
   END IF
   isoFamilies = isotopes(:)%name

   !--- Convection / boundary layer activation for all tracers
   IF(.NOT.ALLOCATED(conv_flg)) ALLOCATE(conv_flg(nbtr)); conv_flg(1:nbtr) = 1
   IF(.NOT.ALLOCATED( pbl_flg)) ALLOCATE( pbl_flg(nbtr));  pbl_flg(1:nbtr) = 1

   !--- Note: nqtottr can differ from nbtr when nmom/=0
   nqtottr = nqtot - COUNT(delPhase(tracers%gen0Name) == 'H2O' .AND. tracers%component == 'lmdz')
   IF(COUNT(tracers%iso_iName == 0) - COUNT(delPhase(tracers%name) == 'H2O' .AND. tracers%component == 'lmdz') /= nqtottr) &
      CALL abort_gcm(modname, 'problem with the computation of nqtottr', 1)

   !=== DISPLAY THE RESULTS
   IF(.NOT.is_master) RETURN
   CALL msg('nqo    = '//TRIM(int2str(nqo)),    modname)
   CALL msg('nbtr   = '//TRIM(int2str(nbtr)),   modname)
   CALL msg('nqtrue = '//TRIM(int2str(nqtrue)), modname)
   CALL msg('nqtot  = '//TRIM(int2str(nqtot)),  modname)
   CALL msg('niso   = '//TRIM(int2str(niso)),   modname)
   CALL msg('ntiso  = '//TRIM(int2str(ntiso)),  modname)
#ifdef INCA
   CALL msg('nqCO2  = '//TRIM(int2str(nqCO2)),  modname)
   CALL msg('nqINCA = '//TRIM(int2str(nqINCA)), modname)
#endif
   t => tracers
   CALL msg('Information stored in '//TRIM(modname)//': ', modname)
   IF(dispTable('isssssssiiiiiiiii', ['iq  ', 'name', 'lNam', 'g0Nm', 'prnt', 'type', 'phas', 'comp',      &
                              'iAdv', 'iGen', 'iqPr', 'nqDe', 'nqCh', 'iGrp', 'iNam', 'iZon', 'iPha'],     &
      cat(t%name, t%longName, t%gen0Name, t%parent, t%type, t%phase, t%component),                         &
      cat([(iq, iq=1, nqtot)], t%iadv, t%iGeneration, t%iqParent, t%nqDescen, t%nqChildren, t%iso_iGroup,  &
                  t%iso_iName, t%iso_iZone, t%iso_iPhase), nColMax=maxTableWidth, nHead=2, sub=modname))   &
      CALL abort_gcm(modname, "problem with the tracers table content", 1)
   CALL msg('No isotopes identified.', modname, nbIso == 0)
   IF(nbIso == 0) RETURN
   DO it = 1, nbIso
      IF(isoSelect(it, .TRUE.)) CALL abort_gcm(modname, 'Problem when selecting isotopes class', 1)
      CALL msg('For isotopes family "'//TRIM(isoFamilies(it))//'":', modname)
      CALL msg('  isoName = '//strStack(isotope%trac),  modname)
      CALL msg('  isoZone = '//strStack(isotope%zone),  modname)
      CALL msg('  isoPhas = '//    TRIM(isotope%phase), modname)
   END DO
   IF(isoSelect('H2O', .TRUE.)) THEN
      IF(isoSelect(1,  .TRUE.)) CALL abort_gcm(modname, 'Problem when selecting isotopes class', 1)
   ELSE
      iH2O = ixIso
   END IF
   IF(ALLOCATED(isotope%keys(ixIso)%key)) &
     CALL msg('  isoKeys('//TRIM(int2str(ixIso))//') = '//TRIM(strStack(isotope%keys(ixIso)%key)), modname)

END SUBROUTINE init_infotrac

!==============================================================================================================================
LOGICAL FUNCTION isoSelectByName(iClass, lVerbose) RESULT(lerr)
   IMPLICIT NONE
   CHARACTER(LEN=*),  INTENT(IN) :: iClass
   LOGICAL, OPTIONAL, INTENT(IN) :: lVerbose
   INTEGER :: iIso
   LOGICAL :: lV
   lV = .FALSE.; IF(PRESENT(lVerbose)) lV = lVerbose
   iIso = strIdx(isotopes(:)%name, iClass)
   lerr = iIso == 0
   IF(lerr) THEN
      niso = 0; ntiso = 0; nzone = 0; nphas = 0; isoCheck=.FALSE.
      CALL msg('no isotope family named "'//TRIM(iClass)//'"', ll=lV)
      RETURN
   END IF
   lerr = isoSelectByIndex(iIso, lV)
END FUNCTION isoSelectByName
!==============================================================================================================================
LOGICAL FUNCTION isoSelectByIndex(iIso, lVerbose) RESULT(lerr)
   IMPLICIT NONE
   INTEGER,           INTENT(IN) :: iIso
   LOGICAL, OPTIONAL, INTENT(IN) :: lVerbose
   LOGICAL :: lV
   lV = .FALSE.;  IF(PRESENT(lVerbose)) lV = lVerbose
   lerr = .FALSE.
   IF(iIso == ixIso) RETURN                                          !--- Nothing to do if the index is already OK
   lerr = iIso<=0 .OR. iIso>SIZE(isotopes)
   CALL msg('Inconsistent isotopes family index '//TRIM(int2str(iIso))//': should be > 0 and <= '&
          //TRIM(int2str(SIZE(isotopes)))//'"', ll = lerr .AND. lV)
   IF(lerr) RETURN
   ixIso = iIso                                                      !--- Update currently selected family index
   isotope  => isotopes(ixIso)                                       !--- Select corresponding component
   isoKeys  => isotope%keys;     niso     = isotope%niso
   isoName  => isotope%trac;     ntiso    = isotope%ntiso
   isoZone  => isotope%zone;     nzone    = isotope%nzone
   isoPhas  => isotope%phase;    nphas    = isotope%nphas
   itZonIso => isotope%itZonIso; isoCheck = isotope%check
   iqIsoPha => isotope%iqIsoPha
END FUNCTION isoSelectByIndex
!==============================================================================================================================

END MODULE infotrac
